import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# Define affine transformation function
def affine_transform(points, scale, rotation, translation):
    """
    Apply scaling, rotation, and translation transformations to a set of points.
    """
    # Scaling matrix
    S = np.diag(scale)
    
    # Rotation matrices
    theta_x, theta_y, theta_z = np.radians(rotation)
    Rx = np.array([[1, 0, 0],
                   [0, np.cos(theta_x), -np.sin(theta_x)],
                   [0, np.sin(theta_x), np.cos(theta_x)]])
    
    Ry = np.array([[np.cos(theta_y), 0, np.sin(theta_y)],
                   [0, 1, 0],
                   [-np.sin(theta_y), 0, np.cos(theta_y)]])
    
    Rz = np.array([[np.cos(theta_z), -np.sin(theta_z), 0],
                   [np.sin(theta_z), np.cos(theta_z), 0],
                   [0, 0, 1]])
    
    R = Rx @ Ry @ Rz  # Combined rotation
    
    # Apply transformations
    transformed_points = (S @ (R @ points.T)).T + translation
    
    return transformed_points

# Generate tree recursively
def generate_tree(points, depth, max_depth):
    """
    Recursively generate a tree structure by applying affine transformations.
    """
    if depth >= max_depth:
        return []
    
    branches = []
    transformations = [
        ((0.3, 0.3, 0.7), (20, 0, 20), [0, 0, 1]),
        ((0.4, 0.3, 0.3), (-10, 0, -10), [0, 0, 1]),
        ((0.3, 0.4, 0.7), (45, 0, 45), [0, 0, 1]),
        ((0.6, 0.4, 0.3), (-25, 0, -25), [0, 0, 1])
    ]
    
    for scale, rotation, translation in transformations:
        new_branch = affine_transform(points, scale, rotation, translation)
        branches.append(new_branch)
        branches.extend(generate_tree(new_branch, depth + 1, max_depth))
    
    return branches

# Define initial trunk
trunk = np.array([[0, 0, 0], [0, 0, 1]])  # A vertical segment

# Generate the tree
max_depth = 7  # Number of recursive iterations
tree_branches = generate_tree(trunk, 0, max_depth)

# Plot the tree
fig = plt.figure(figsize=(8, 8))
ax = fig.add_subplot(111, projection='3d')
ax.set_xlim([-2, 2])
ax.set_ylim([-2, 2])
ax.set_zlim([0, 5])
ax.set_xlabel("X")
ax.set_ylabel("Y")
ax.set_zlabel("Z")
ax.set_title("Fractal Tree Using Iterated Function System (IFS)")

# Plot branches
for branch in tree_branches:
    ax.plot(branch[:, 0], branch[:, 1], branch[:, 2], 'brown')

plt.show()
