import numpy as np
import os

from cce_alg import evolution
from sce_plots import generate_plots, gif, plot_loss, plot_3D
from sce_helper import sort, partition

# This function implements the Shuffled Complex Evolution (SCE) algorithm 
# as described in https://doi.org/10.1016/j.ecolmodel.2010.10.002
def sce(parameters, cost_function, output_prefix):

    # Unpack the parameters for the SCE
    population_size, criteria, iter, partition_amount, alpha, beta = parameters

    # Initialize a random population
    population = np.random.uniform(low=-100, high=100, size=(population_size, 2)).astype(np.float64)
    
    # Setup arrays to save the population and the loss at each step (and at the end) 
    iter_best_loss = []
    iter_population = []

    # Each iteration is a step in the Shuffled Complex Evolution algorithm
    for iteration in range(iter):
        
        # Sort the population according to the given cost (or loss) function
        sorted_population, _ = sort(population, cost=cost_function)

        # Save the current population and the current best candidate's cost
        iter_population.append(sorted_population)
        iter_best_loss.append(cost_function(sorted_population[0][0], sorted_population[0][1]))
        
        print(iteration, "Best individual", sorted_population[0], "Cost:", cost_function(sorted_population[0][0], sorted_population[0][1]))
        
        # If a convergence criteria is defined, if the current best solution verifies it, we stop the SCE algorithm
        if criteria != None and cost_function(sorted_population[0][0], sorted_population[0][1]) < criteria:
            print("Criteria satisfied by individual", sorted_population[0], 
                  "with cost", cost_function(sorted_population[0][0], sorted_population[0][1]),
                  "at iteration", iteration)
            break

        # Partition the population in communities, according to the parameters given to the SCE algorithm
        partitionned_population = partition(sorted_population,partition_amount, show_warning=True)

        # Generate the new population using the Competitive Complex Evolution algorithm, applied on each community separately
        offsprings_partition = evolution(partitionned_population, population_size, partition_amount, alpha, beta, cost_function)
        
        # Break up the communities and put all individuals back together
        population = np.concatenate(offsprings_partition)
        
        # If this is the last iteration, save the last generated population and its best loss
        if iteration == iter - 1:
            sorted_population, _ = sort(population, cost=cost_function)
            iter_population.append(sorted_population)
            iter_best_loss.append(cost_function(sorted_population[0][0], sorted_population[0][1]))
            print(iteration+1, "Best individual", sorted_population[0], "Cost:", cost_function(sorted_population[0][0], sorted_population[0][1]))
    
    # Give the best candidate found over all iterations (not necessarely the last one)
    global_min_index, global_min_value = min(enumerate(iter_best_loss), key=lambda x: x[1])
    print("Best found over", iter, "iterations is", global_min_value, "at iteration", global_min_index)

    # Define the output folders and gif names
    population_folder = os.path.join('results', output_prefix, output_prefix + '_population_images')
    radius_folder = os.path.join('results', output_prefix, output_prefix + '_radius_images')
    population_gif_name = os.path.join('results', output_prefix, output_prefix + '_population.gif')
    radius_gif_name = os.path.join('results', output_prefix, output_prefix + '_radius.gif')
    cost_plot_filename = os.path.join('results', output_prefix, output_prefix + '_cost.png')
    plot_3D_filename = os.path.join('results', output_prefix, output_prefix + '_3d_plot.gif')

    # Make sure the directories to save the output images exist
    os.makedirs('results', exist_ok=True)
    os.makedirs(os.path.join('results', output_prefix), exist_ok=True)
    os.makedirs(population_folder, exist_ok=True)
    os.makedirs(radius_folder, exist_ok=True)

    # Plot the distribution of individuals for each step of the algorithm
    print("Generating plots...")
    for iteration in range(iter):            
        generate_plots(iter_population[iteration], iteration, population_folder, radius_folder)

    # Make a GIF file out of the generated plots
    gif(population_folder, population_gif_name, iter)
    gif(radius_folder, radius_gif_name, iter)

    # Plot the best cost as a function of the iterations
    plot_loss(iter_best_loss, cost_plot_filename)
    
    # Plot the distribution of the population for each step, projected onto the 3D plot of the cost function 
    plot_3D(iter_population, cost_function, plot_3D_filename)


# Goldstein-Price cost function, modified so that the minimum has a value of 0 instead of 3
def goldstein_price(x, y):
    return abs(3 - (1 + (x + y + 1)**2 * (19 - 14*x + 3*x**2 - 14*y + 6*x*y + 3*y**2)) * (30 + (2*x - 3*y)**2 * (18 - 32*x + 12*x**2 + 48*y - 36*x*y + 27*y**2)))

def squared(x, y):
    return abs(x**2+y**2)

def rosenbrock(x, y):
    return abs((1 - x)**2 + 100 * (y - x**2)**2)

def main():

    # Define the prefix used in folder names for the population and population distribution plots
    output_prefix = 'goldstein_price'
    
    # Parameters used in the Shuffled Complex Evolution and Competitive Complex Evolution algorithms
    # Population size, Criteria to stop algorithm, Maximum iterations, Partition size, Alpha, Beta  
    sce_parameters = (20000, None, 150, 2000, 2, 2)

    # Start the SCE algorithm with the given parameters, the cost function and the output folders
    sce(sce_parameters, goldstein_price, output_prefix)
    
    # Examples for other cost functions
    output_prefix = 'squared'
    sce(sce_parameters, squared, output_prefix)

    output_prefix = 'rosenbrock'
    sce(sce_parameters, rosenbrock, output_prefix)

main()
