import imageio
import matplotlib.pyplot as plt
import matplotlib.animation as anim
from scipy.spatial.distance import cdist
import numpy as np
import os

def plot_population(population, iteration, output_folder):
    plt.figure(figsize=(6, 6))
    plt.scatter(population[:, 0], population[:, 1], c='blue', marker='o')
    plt.title(f'Iteration {iteration}')
    plt.xlim(-110, 110)
    plt.ylim(-110, 110)
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.grid(True)
    plt.savefig(os.path.join(output_folder, f'iteration_{iteration}.png'))
    plt.close()

def plot_radii(iteration, output_folder, barycenter, radii):
    plt.figure(figsize=(6, 6))
    plt.scatter(barycenter[0], barycenter[1], c='red', marker='x', s=100, label='Barycenter')

    for radius, label, color in zip(radii, ['50%', '95%', '99.9%'], ['g', 'b', 'r']):
        circle = plt.Circle(barycenter, radius, color=color, fill=False, linestyle='--', label=f'{label} Circle')
        plt.gca().add_patch(circle)

    plt.title(f'Iteration {iteration}')
    plt.xlim(-100, 100)
    plt.ylim(-100, 100)
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.grid(True)
    plt.legend()
    plt.savefig(os.path.join(output_folder, f'iteration_{iteration}.png'))
    plt.close()

def calculate_barycenter(population):
    return np.mean(population, axis=0)

def calculate_radii(population, barycenter, percentiles):
    distances = cdist(population, [barycenter]).flatten()
    return np.percentile(distances, percentiles)

def gif(images_folder, filepath, iterations):
    images = []
    for iteration in range(iterations):
        image_path = os.path.join(images_folder, f'iteration_{iteration}.png')
        images.append(imageio.imread(image_path))
    imageio.mimsave(filepath, images, duration=0.5)

def generate_plots(population, iteration, population_folder, radius_folder):
    barycenter = calculate_barycenter(population)

    radii = calculate_radii(population, barycenter, [50, 95, 99.9])

    plot_population(population, iteration, population_folder)

    plot_radii(iteration, radius_folder, barycenter, radii)

def plot_loss(loss_values, filepath):
    # Create a plot
    plt.figure(figsize=(10, 6))
    plt.plot(loss_values, marker='o', linestyle='-', color='b')

    # Set labels and title
    plt.xlabel('Iteration Number')
    plt.ylabel('Loss')
    plt.title('Loss per Iteration')
    plt.yscale('log')

    # Show the plot
    plt.grid(True)
    plt.savefig(filepath)
    plt.close()

def plot_3D(populations, function_to_plot, filepath):
    populations = np.array(populations)
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111, projection='3d')
    
    # Define meshgrid for surface plot
    x = np.linspace(np.min(populations[..., 0]), np.max(populations[..., 0]), 100)
    y = np.linspace(np.min(populations[..., 1]), np.max(populations[..., 1]), 100)
    X, Y = np.meshgrid(x, y)
    Z = function_to_plot(X, Y)
    
    ax.plot_surface(X, Y, Z, cmap='viridis', alpha=0.6)
    
    # Scatter points
    scatter = ax.scatter([], [], [], color='red', s=1)
    
    def update(frame):
        ax.set_title(f'Iteration {frame}')
        points = populations[frame]
        Z_points = function_to_plot(points[:, 0], points[:, 1])
        scatter._offsets3d = (points[:, 0], points[:, 1], Z_points)
        return scatter,
    
    ani = anim.FuncAnimation(fig, update, frames=len(populations), interval=50, blit=False)
    ani.save(filepath, writer='pillow', fps=20)


